@@ -1,29 +1,34 @@
 #!/usr/bin/env python
 # coding=utf8
 # This file is part of Xpra.
-# Copyright (C) 2013, 2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2013 - 2016 Antoine Martin <antoine@devloop.org.uk>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-import sys
+import sys, os
 from threading import Lock
 from xpra.log import Logger
 log = Logger("codec", "video")
 
-from xpra.codecs.loader import get_codec, get_codec_error, load_codecs
+from xpra.codecs.loader import get_codec, get_codec_error
+from xpra.util import csv, engs
 
 
+NVENC_OPTIONS = [("nvenc%s" % v) for v in os.environ.get("XPRA_NVENC_VERSIONS", "6,5,4").split(",")]
+
 #the codec loader uses the names...
 #but we need the module name to be able to probe without loading the codec:
-CODEC_TO_MODULE = {"vpx"        : "vpx",
-                   "x264"       : "enc_x264",
-                   "x265"       : "enc_x265",
-                   "nvenc"      : "nvenc",
-                   "swscale"    : "csc_swscale",
-                   "cython"     : "csc_cython",
-                   "opencl"     : "csc_opencl",
-                   "avcodec"    : "dec_avcodec",
-                   "avcodec2"   : "dec_avcodec2"}
+CODEC_TO_MODULE = {"vpx"        : ["vpx"],
+                   "x264"       : ["enc_x264"],
+                   "x265"       : ["enc_x265"],
+                   "nvenc"      : NVENC_OPTIONS,
+                   "xvid"       : ["xvid"],
+                   "swscale"    : ["csc_swscale"],
+                   "cython"     : ["csc_cython"],
+                   "opencl"     : ["csc_opencl"],
+                   "opencv"     : ["csc_opencv"],
+                   "libyuv"     : ["csc_libyuv"],
+                   "avcodec2"   : ["dec_avcodec2"]}
 
 def has_codec_module(module_name):
     top_module = "xpra.codecs.%s" % module_name
@@ -31,27 +36,29 @@ def has_codec_module(module_name):
         __import__(top_module, {}, {}, [])
         log("codec module %s is installed", module_name)
         return True
-    except Exception, e:
+    except Exception as e:
         log("codec module %s cannot be loaded: %s", module_name, e)
         return False
 
 def try_import_modules(codec_names):
     names = []
     for codec_name in codec_names:
-        module_name = CODEC_TO_MODULE[codec_name]
-        if has_codec_module(module_name):
-            names.append(codec_name)
+        module_names = CODEC_TO_MODULE[codec_name]
+        for module_name in module_names:
+            if has_codec_module(module_name):
+                names.append(codec_name)
+                break
     return names
 
 #all the codecs we know about:
 #try to import the module that contains them (cheap check):
-ALL_VIDEO_ENCODER_OPTIONS = try_import_modules(["x264", "vpx", "x265", "nvenc"])
-ALL_CSC_MODULE_OPTIONS = try_import_modules(["swscale", "cython", "opencl"])
+ALL_VIDEO_ENCODER_OPTIONS = try_import_modules(["x264", "vpx", "xvid", "x265", "nvenc"])
+ALL_CSC_MODULE_OPTIONS = try_import_modules(["swscale", "cython", "opencl", "libyuv", "opencv"])
 NO_GFX_CSC_OPTIONS = [x for x in ALL_CSC_MODULE_OPTIONS if x not in ("opencl", )]
-ALL_VIDEO_DECODER_OPTIONS = try_import_modules(["avcodec", "avcodec2", "vpx"])
+ALL_VIDEO_DECODER_OPTIONS = try_import_modules(["avcodec2", "vpx"])
 
-PREFERRED_ENCODER_ORDER = ["nvenc", "x264", "vpx", "x265"]
-PREFERRED_DECODER_ORDER = ["avcodec", "avcodec2", "vpx"]
+PREFERRED_ENCODER_ORDER = ["nvenc", "x264", "vpx", "xvid", "x265"]
+PREFERRED_DECODER_ORDER = ["avcodec2", "vpx"]
 log("video_helper: ALL_VIDEO_ENCODER_OPTIONS=%s", ALL_VIDEO_ENCODER_OPTIONS)
 log("video_helper: ALL_CSC_MODULE_OPTIONS=%s", ALL_CSC_MODULE_OPTIONS)
 log("video_helper: NO_GFX_CSC_OPTIONS=%s", NO_GFX_CSC_OPTIONS)
@@ -61,10 +68,14 @@ log("video_helper: ALL_VIDEO_DECODER_OPTIONS=%s", ALL_VIDEO_DECODER_OPTIONS)
 #on top of that, there are compatibility problems with gtk at times: OpenCL AMD and TLS don't mix well
 
 
-def get_encoder_module_name(x):
-        if x.find("enc")>=0:
-            return x            #ie: "nvenc" or "enc_vpx"
-        return "enc_"+x         #ie: "enc_x264"
+def get_encoder_module_names(x):
+    if x=="nvenc":
+        return NVENC_OPTIONS
+    if x=="xvid":
+        return ["xvid"]
+    elif x.find("enc")>=0:
+        return [x]              #ie: "nvenc" or "enc_vpx"
+    return ["enc_"+x]           #ie: "enc_x264"
 
 def get_decoder_module_name(x):
         return "dec_"+x         #ie: "dec_vpx"
@@ -78,10 +89,12 @@ def get_DEFAULT_VIDEO_ENCODERS():
     """ returns all the video encoders installed """
     encoders = []
     for x in list(ALL_VIDEO_ENCODER_OPTIONS):
-        mod = get_encoder_module_name(x)
-        c = get_codec(mod)
-        if c:
-            encoders.append(x)
+        mods = get_encoder_module_names(x)
+        for mod in mods:
+            c = get_codec(mod)
+            if c:
+                encoders.append(x)
+                break
     return encoders
 
 def get_DEFAULT_CSC_MODULES():
@@ -139,11 +152,10 @@ class VideoHelper(object):
         self.csc_modules    = filt("csc modules"    , csc_modules,      ALL_CSC_MODULE_OPTIONS)
         self.video_decoders = filt("video decoders" , video_decoders,   ALL_VIDEO_DECODER_OPTIONS)
         log("VideoHelper.set_modules(%s, %s, %s) video encoders=%s, csc=%s, video decoders=%s",
-            video_encoders, csc_modules, video_decoders, self.video_encoders, self.csc_modules, self.video_decoders)
+            csv(video_encoders), csv(csc_modules), csv(video_decoders), csv(self.video_encoders), csv(self.csc_modules), csv(self.video_decoders))
 
     def cleanup(self):
-        try:
-            self._lock.acquire()
+        with self._lock:
             #check again with lock held (in case of race):
             if not self._initialized:
                 return
@@ -160,8 +172,6 @@ class VideoHelper(object):
             self.csc_modules = []
             self.video_decoders = []
             self._initialized = False
-        finally:
-            self._lock.release()
 
     def clone(self):
         if not self._initialized:
@@ -182,18 +192,21 @@ class VideoHelper(object):
 
     def get_info(self):
         d = {}
+        einfo = d.setdefault("encoding", {})
+        dinfo = d.setdefault("decoding", {})
+        cinfo = d.setdefault("csc", {})
         for encoding, encoder_specs in self._video_encoder_specs.items():
             for in_csc, specs in encoder_specs.items():
                 for spec in specs:
-                    d.setdefault("encoding."+in_csc+"_to_"+encoding, []).append(spec.codec_type)
+                    einfo.setdefault("%s_to_%s" % (in_csc, encoding), []).append(spec.codec_type)
         for in_csc, specs in self._csc_encoder_specs.items():
             for out_csc, specs in specs.items():
-                d["csc."+in_csc+"_to_"+out_csc] = [spec.codec_type for spec in specs]
+                cinfo["%s_to_%s" % (in_csc, out_csc)] = [spec.codec_type for spec in specs]
         for encoding, decoder_specs in self._video_decoder_specs.items():
             for out_csc, decoders in decoder_specs.items():
                 for decoder in decoders:
                     decoder_name, _ = decoder
-                    d.setdefault("decoding."+encoding+"_to_"+out_csc, []).append(decoder_name)
+                    dinfo.setdefault("%s_to_%s" % (encoding, out_csc), []).append(decoder_name)
         def modstatus(x, def_list, active_list):
             #the module is present
             if x in active_list:
@@ -202,17 +215,17 @@ class VideoHelper(object):
                 return "disabled"
             else:
                 return "not found"
+        venc = einfo.setdefault("video-encoder", {})
         for x in ALL_VIDEO_ENCODER_OPTIONS:
-            d["encoding.video-encoder.%s" % x] = modstatus(x, get_DEFAULT_VIDEO_ENCODERS(), self.video_encoders)
+            venc["%s" % x] = modstatus(x, get_DEFAULT_VIDEO_ENCODERS(), self.video_encoders)
+        cscm = einfo.setdefault("csc-module", {})
         for x in ALL_CSC_MODULE_OPTIONS:
-            d["encoding.csc-module.%s" % x] = modstatus(x, get_DEFAULT_CSC_MODULES(), self.csc_modules)
+            cscm["%s" % x] = modstatus(x, get_DEFAULT_CSC_MODULES(), self.csc_modules)
         return d
 
     def init(self):
         log("VideoHelper.init()")
-        load_codecs()
-        try:
-            self._lock.acquire()
+        with self._lock:
             #check again with lock held (in case of race):
             log("VideoHelper.init() initialized=%s", self._initialized)
             if self._initialized:
@@ -221,8 +234,6 @@ class VideoHelper(object):
             self.init_csc_options()
             self.init_video_decoders_options()
             self._initialized = True
-        finally:
-            self._lock.release()
         log("VideoHelper.init() done")
 
     def get_encodings(self):
@@ -246,37 +257,44 @@ class VideoHelper(object):
 
 
     def init_video_encoders_options(self):
-        log("init_video_encoders_options() will try video encoders: %s", self.video_encoders)
+        log("init_video_encoders_options()")
+        log(" will try video encoders: %s", csv(self.video_encoders))
         for x in self.video_encoders:
             try:
-                mod = get_encoder_module_name(x)
-                self.init_video_encoder_option(mod)
-            except:
-                log.warn("init_video_encoders_options() cannot add %s encoder", x, exc_info=True)
-        log("init_video_encoders_options() video encoder specs: %s", self._video_encoder_specs)
+                mods = get_encoder_module_names(x)
+                log(" modules for %s: %s", x, csv(mods))
+                for mod in mods:
+                    try:
+                        self.init_video_encoder_option(mod)
+                        break
+                    except Exception as e:
+                        log(" init_video_encoder_option(%s) error", mod, exc_info=True)
+                        log.warn("Warning: cannot load %s video encoder:", mod)
+                        log.warn(" %s", e)
+            except Exception as e:
+                log.warn("Warning: cannot add %s encoder: %s", x, e)
+        log("found %i video encoder%s: %s", len(self._video_encoder_specs), engs(self._video_encoder_specs), csv(self._video_encoder_specs))
 
     def init_video_encoder_option(self, encoder_name):
         encoder_module = get_codec(encoder_name)
-        log("init_video_encoder_option(%s) module=%s", encoder_name, encoder_module)
+        log("init_video_encoder_option(%s)", encoder_name)
+        log(" module=%s", encoder_module)
         if not encoder_module:
-            log.warn("video encoder '%s' could not be loaded:", encoder_name)
-            log.warn(" %s", get_codec_error(encoder_name))
+            log(" video encoder '%s' could not be loaded:", encoder_name)
+            log(" %s", get_codec_error(encoder_name))
             return
         encoder_type = encoder_module.get_type()
         try:
             encoder_module.init_module()
             self._cleanup_modules.append(encoder_module)
-        except Exception, e:
-            log("exception in %s module initialization %s: %s", encoder_type, encoder_module.init_module, e, exc_info=True)
-            log.warn("Warning: %s video encoder failed: %s", encoder_type, e)
-            return
-        colorspaces = encoder_module.get_input_colorspaces()
-        log("init_video_encoder_option(%s) %s input colorspaces=%s", encoder_module, encoder_type, colorspaces)
+        except Exception as e:
+            log(" exception in %s module %s initialization %s: %s", encoder_type, encoder_module.__name__, encoder_module.init_module, e, exc_info=True)
+            raise
         encodings = encoder_module.get_encodings()
-        log("init_video_encoder_option(%s) %s encodings=%s", encoder_module, encoder_type, encodings)
+        log(" %s encodings=%s", encoder_type, csv(encodings))
         for encoding in encodings:
-            #we assume that all encodings support all colorspaces here...
-            #which may not be the case in the future!
+            colorspaces = encoder_module.get_input_colorspaces(encoding)
+            log(" %s input colorspaces for %s: %s", encoder_type, encoding, csv(colorspaces))
             for colorspace in colorspaces:
                 spec = encoder_module.get_spec(encoding, colorspace)
                 self.add_encoder_spec(encoding, colorspace, spec)
@@ -286,37 +304,40 @@ class VideoHelper(object):
 
 
     def init_csc_options(self):
-        log("init_csc_options() will try csc modules: %s", self.csc_modules)
+        log("init_csc_options()")
+        log(" will try csc modules: %s", csv(self.csc_modules))
         for x in self.csc_modules:
             try:
                 mod = get_csc_module_name(x)
                 self.init_csc_option(mod)
             except:
                 log.warn("init_csc_options() cannot add %s csc", x, exc_info=True)
-        log("init_csc_options() csc specs: %s", self._csc_encoder_specs)
+        log(" csc specs: %s", csv(self._csc_encoder_specs))
         for src_format, d in sorted(self._csc_encoder_specs.items()):
-            log("%s - %s options:", src_format, len(d))
+            log(" %s - %s options:", src_format, len(d))
             for dst_format, specs in sorted(d.items()):
-                log(" * %s via: %s", dst_format, sorted(list(specs), key=lambda spec: spec.codec_type))
+                log("  * %s via: %s", dst_format, csv(sorted(spec.codec_type for spec in specs)))
 
     def init_csc_option(self, csc_name):
         csc_module = get_codec(csc_name)
-        log("init_csc_option(%s) module=%s", csc_name, csc_module)
+        log("init_csc_option(%s)", csc_name)
+        log(" module=%s", csc_module)
         if csc_module is None:
-            log.warn("csc module %s could not be loaded: %s", csc_name, get_codec_error(csc_name))
+            log(" csc module %s could not be loaded:", csc_name)
+            log(" %s", get_codec_error(csc_name))
             return
         csc_type = csc_module.get_type()
         try:
             csc_module.init_module()
             self._cleanup_modules.append(csc_module)
-        except Exception, e:
+        except Exception as e:
             log("exception in %s module initialization %s: %s", csc_type, csc_module.init_module, e, exc_info=True)
-            log.warn("cannot use %s module %s: %s", csc_type, csc_module, e)
+            log.warn("Warning: cannot use %s module %s: %s", csc_type, csc_module, e)
             return
         in_cscs = csc_module.get_input_colorspaces()
         for in_csc in in_cscs:
             out_cscs = csc_module.get_output_colorspaces(in_csc)
-            log("init_csc_option(..) %s.get_output_colorspaces(%s)=%s", csc_module.get_type(), in_csc, out_cscs)
+            log("%s output colorspaces for %s: %s", csc_module.get_type(), in_csc, csv(out_cscs))
             for out_csc in out_cscs:
                 spec = csc_module.get_spec(in_csc, out_csc)
                 self.add_csc_spec(in_csc, out_csc, spec)
@@ -326,41 +347,44 @@ class VideoHelper(object):
 
 
     def init_video_decoders_options(self):
-        log("init_video_decoders_options() will try video decoders: %s", self.video_decoders)
+        log("init_video_decoders_options()")
+        log(" will try video decoders: %s", csv(self.video_decoders))
         for x in self.video_decoders:
             try:
                 mod = get_decoder_module_name(x)
                 self.init_video_decoder_option(mod)
             except:
-                log.warn("init_video_decoders_options() cannot add %s decoder", x, exc_info=True)
-        log("init_video_decoders_options() video decoder specs: %s", self._video_decoder_specs)
+                log.warn("Warning: cannot add %s decoder", x, exc_info=True)
+        log("found %s video decoder%s: %s", len(self._video_decoder_specs), engs(self._video_decoder_specs), csv(self._video_decoder_specs))
 
     def init_video_decoder_option(self, decoder_name):
         decoder_module = get_codec(decoder_name)
-        log("init_video_decoder_option(%s) module=%s", decoder_name, decoder_module)
+        log("init_video_decoder_option(%s)", decoder_name)
+        log(" module=%s", decoder_module)
         if not decoder_module:
-            log.warn("video decoder %s could not be loaded: %s", decoder_name, get_codec_error(decoder_name))
+            log(" video decoder %s could not be loaded:", decoder_name)
+            log(" %s", get_codec_error(decoder_name))
             return
-        encoder_type = decoder_module.get_type()
+        decoder_type = decoder_module.get_type()
         try:
             decoder_module.init_module()
             self._cleanup_modules.append(decoder_module)
-        except Exception, e:
-            log("exception in %s module initialization %s: %s", encoder_type, decoder_module.init_module, e, exc_info=True)
-            log.warn("cannot use %s module %s: %s", encoder_type, decoder_module, e, exc_info=True)
+        except Exception as e:
+            log("exception in %s module initialization %s: %s", decoder_type, decoder_module.init_module, e, exc_info=True)
+            log.warn("Warning: cannot use %s module %s: %s", decoder_type, decoder_module, e, exc_info=True)
             return
         encodings = decoder_module.get_encodings()
-        log("init_video_decoder_option(%s) %s encodings=%s", decoder_module, encoder_type, encodings)
+        log(" %s encodings=%s", decoder_type, csv(encodings))
         for encoding in encodings:
             colorspaces = decoder_module.get_input_colorspaces(encoding)
-            log("init_video_decoder_option(%s) %s input colorspaces for %s: %s", decoder_module, encoder_type, encoding, colorspaces)
+            log(" %s input colorspaces for %s: %s", decoder_type, encoding, csv(colorspaces))
             for colorspace in colorspaces:
                 output_colorspace = decoder_module.get_output_colorspace(encoding, colorspace)
-                log("init_video_decoder_option(%s) get_output_colorspace(%s, %s)=%s", decoder_name, encoding, colorspace, output_colorspace)
+                log(" %s output colorspace for %s/%s: %s", decoder_type, encoding, colorspace, output_colorspace)
                 try:
                     assert decoder_module.Decoder
                     self.add_decoder(encoding, colorspace, decoder_name, decoder_module)
-                except Exception, e:
+                except Exception as e:
                     log.warn("failed to add decoder %s: %s", decoder_module, e)
 
     def add_decoder(self, encoding, colorspace, decoder_name, decoder_module):
@@ -412,18 +436,22 @@ def getVideoHelper():
 
 
 def main():
-    from xpra.codecs.loader import log as loader_log
-    loader_log.enable_debug()
-    if "-v" in sys.argv or "--verbose" in sys.argv:
-        log.enable_debug()
-    vh = getVideoHelper()
-    vh.set_modules(ALL_VIDEO_ENCODER_OPTIONS, ALL_CSC_MODULE_OPTIONS, ALL_VIDEO_DECODER_OPTIONS)
-    vh.init()
-    log.info("VideoHelper.get_info():")
-    info = vh.get_info()
-    for k in sorted(info.keys()):
-        v = info.get(k)
-        log.info("%s=%s", k, v)
+    from xpra.codecs.loader import log as loader_log, load_codecs
+    from xpra.util import print_nested_dict
+    from xpra.log import enable_color
+    from xpra.platform import program_context
+    with program_context("Video Helper"):
+        enable_color()
+        if "-v" in sys.argv or "--verbose" in sys.argv:
+            loader_log.enable_debug()
+            log.enable_debug()
+        load_codecs()
+        vh = getVideoHelper()
+        vh.set_modules(ALL_VIDEO_ENCODER_OPTIONS, ALL_CSC_MODULE_OPTIONS, ALL_VIDEO_DECODER_OPTIONS)
+        vh.init()
+        log.info("VideoHelper.get_info():")
+        info = vh.get_info()
+        print_nested_dict(info)
 
 
 if __name__ == "__main__":
